@@ -1,13 +1,12 @@
 use rustc_abi::Size;
+use rustc_target::spec::Os;
 
-use crate::concurrency::sync::LAZY_INIT_COOKIE;
+use crate::concurrency::sync::{AccessKind, SyncObj};
 use crate::*;
 
-/// Do a bytewise comparison of the two places, using relaxed atomic reads. This is used to check if
+/// Do a bytewise comparison of the two places. This is used to check if
 /// a synchronization primitive matches its static initializer value.
-///
-/// The reads happen in chunks of 4, so all racing accesses must also use that access size.
-fn bytewise_equal_atomic_relaxed<'tcx>(
+fn bytewise_equal<'tcx>(
     ecx: &MiriInterpCx<'tcx>,
     left: &MPlaceTy<'tcx>,
     right: &MPlaceTy<'tcx>,
@@ -15,33 +14,24 @@ fn bytewise_equal_atomic_relaxed<'tcx>(
     let size = left.layout.size;
     assert_eq!(size, right.layout.size);
 
-    // We do this in chunks of 4, so that we are okay to race with (sufficiently aligned)
-    // 4-byte atomic accesses.
-    assert!(size.bytes().is_multiple_of(4));
-    for i in 0..(size.bytes() / 4) {
-        let offset = Size::from_bytes(i.strict_mul(4));
-        let load = |place: &MPlaceTy<'tcx>| {
-            let byte = place.offset(offset, ecx.machine.layouts.u32, ecx)?;
-            ecx.read_scalar_atomic(&byte, AtomicReadOrd::Relaxed)?.to_u32()
-        };
-        let left = load(left)?;
-        let right = load(right)?;
-        if left != right {
-            return interp_ok(false);
-        }
-    }
+    let left_bytes = ecx.read_bytes_ptr_strip_provenance(left.ptr(), size)?;
+    let right_bytes = ecx.read_bytes_ptr_strip_provenance(right.ptr(), size)?;
 
-    interp_ok(true)
+    interp_ok(left_bytes == right_bytes)
 }
 
+// The in-memory marker values we use to indicate whether objects have been initialized.
+const PTHREAD_UNINIT: u8 = 0;
+const PTHREAD_INIT: u8 = 1;
+
 // # pthread_mutexattr_t
 // We store some data directly inside the type, ignoring the platform layout:
 // - kind: i32
 
 #[inline]
 fn mutexattr_kind_offset<'tcx>(ecx: &MiriInterpCx<'tcx>) -> InterpResult<'tcx, u64> {
-    interp_ok(match &*ecx.tcx.sess.target.os {
-        "linux" | "illumos" | "solaris" | "macos" | "freebsd" | "android" => 0,
+    interp_ok(match &ecx.tcx.sess.target.os {
+        Os::Linux | Os::Illumos | Os::Solaris | Os::MacOs | Os::FreeBsd | Os::Android => 0,
         os => throw_unsup_format!("`pthread_mutexattr` is not supported on {os}"),
     })
 }
@@ -103,7 +93,7 @@ fn mutexattr_translate_kind<'tcx>(
 
 // # pthread_mutex_t
 // We store some data directly inside the type, ignoring the platform layout:
-// - init: u32
+// - init: u8
 
 /// The mutex kind.
 #[derive(Debug, Clone, Copy)]
@@ -120,14 +110,29 @@ struct PthreadMutex {
     kind: MutexKind,
 }
 
+impl SyncObj for PthreadMutex {
+    fn on_access<'tcx>(&self, access_kind: AccessKind) -> InterpResult<'tcx> {
+        if !self.mutex_ref.queue_is_empty() {
+            throw_ub_format!(
+                "{access_kind} of `pthread_mutex_t` is forbidden while the queue is non-empty"
+            );
+        }
+        interp_ok(())
+    }
+
+    fn delete_on_write(&self) -> bool {
+        true
+    }
+}
+
 /// To ensure an initialized mutex that was moved somewhere else can be distinguished from
 /// a statically initialized mutex that is used the first time, we pick some offset within
 /// `pthread_mutex_t` and use it as an "initialized" flag.
 fn mutex_init_offset<'tcx>(ecx: &MiriInterpCx<'tcx>) -> InterpResult<'tcx, Size> {
-    let offset = match &*ecx.tcx.sess.target.os {
-        "linux" | "illumos" | "solaris" | "freebsd" | "android" => 0,
+    let offset = match &ecx.tcx.sess.target.os {
+        Os::Linux | Os::Illumos | Os::Solaris | Os::FreeBsd | Os::Android => 0,
         // macOS stores a signature in the first bytes, so we move to offset 4.
-        "macos" => 4,
+        Os::MacOs => 4,
         os => throw_unsup_format!("`pthread_mutex` is not supported on {os}"),
     };
     let offset = Size::from_bytes(offset);
@@ -138,23 +143,23 @@ fn mutex_init_offset<'tcx>(ecx: &MiriInterpCx<'tcx>) -> InterpResult<'tcx, Size>
         let check_static_initializer = |name| {
             let static_initializer = ecx.eval_path(&["libc", name]);
             let init_field =
-                static_initializer.offset(offset, ecx.machine.layouts.u32, ecx).unwrap();
-            let init = ecx.read_scalar(&init_field).unwrap().to_u32().unwrap();
-            assert_ne!(
-                init, LAZY_INIT_COOKIE,
-                "{name} is incompatible with our initialization cookie"
+                static_initializer.offset(offset, ecx.machine.layouts.u8, ecx).unwrap();
+            let init = ecx.read_scalar(&init_field).unwrap().to_u8().unwrap();
+            assert_eq!(
+                init, PTHREAD_UNINIT,
+                "{name} is incompatible with our initialization logic"
             );
         };
 
         check_static_initializer("PTHREAD_MUTEX_INITIALIZER");
         // Check non-standard initializers.
-        match &*ecx.tcx.sess.target.os {
-            "linux" => {
+        match &ecx.tcx.sess.target.os {
+            Os::Linux => {
                 check_static_initializer("PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP");
                 check_static_initializer("PTHREAD_ERRORCHECK_MUTEX_INITIALIZER_NP");
                 check_static_initializer("PTHREAD_ADAPTIVE_MUTEX_INITIALIZER_NP");
             }
-            "illumos" | "solaris" | "macos" | "freebsd" | "android" => {
+            Os::Illumos | Os::Solaris | Os::MacOs | Os::FreeBsd | Os::Android => {
                 // No non-standard initializers.
             }
             os => throw_unsup_format!("`pthread_mutex` is not supported on {os}"),
@@ -172,7 +177,7 @@ fn mutex_create<'tcx>(
 ) -> InterpResult<'tcx, PthreadMutex> {
     let mutex = ecx.deref_pointer_as(mutex_ptr, ecx.libc_ty_layout("pthread_mutex_t"))?;
     let data = PthreadMutex { mutex_ref: MutexRef::new(), kind };
-    ecx.lazy_sync_init(&mutex, mutex_init_offset(ecx)?, data.clone())?;
+    ecx.init_immovable_sync(&mutex, mutex_init_offset(ecx)?, PTHREAD_INIT, data.clone())?;
     interp_ok(data)
 }
 
@@ -186,10 +191,11 @@ where
     'tcx: 'a,
 {
     let mutex = ecx.deref_pointer_as(mutex_ptr, ecx.libc_ty_layout("pthread_mutex_t"))?;
-    ecx.lazy_sync_get_data(
+    ecx.get_immovable_sync_with_static_init(
         &mutex,
         mutex_init_offset(ecx)?,
-        || throw_ub_format!("`pthread_mutex_t` can't be moved after first use"),
+        PTHREAD_UNINIT,
+        PTHREAD_INIT,
         |ecx| {
             let kind = mutex_kind_from_static_initializer(ecx, &mutex)?;
             interp_ok(PthreadMutex { mutex_ref: MutexRef::new(), kind })
@@ -203,16 +209,15 @@ fn mutex_kind_from_static_initializer<'tcx>(
     mutex: &MPlaceTy<'tcx>,
 ) -> InterpResult<'tcx, MutexKind> {
     // All the static initializers recognized here *must* be checked in `mutex_init_offset`!
-    let is_initializer =
-        |name| bytewise_equal_atomic_relaxed(ecx, mutex, &ecx.eval_path(&["libc", name]));
+    let is_initializer = |name| bytewise_equal(ecx, mutex, &ecx.eval_path(&["libc", name]));
 
     // PTHREAD_MUTEX_INITIALIZER is recognized on all targets.
     if is_initializer("PTHREAD_MUTEX_INITIALIZER")? {
         return interp_ok(MutexKind::Default);
     }
     // Support additional platform-specific initializers.
-    match &*ecx.tcx.sess.target.os {
-        "linux" =>
+    match &ecx.tcx.sess.target.os {
+        Os::Linux =>
             if is_initializer("PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP")? {
                 return interp_ok(MutexKind::Recursive);
             } else if is_initializer("PTHREAD_ERRORCHECK_MUTEX_INITIALIZER_NP")? {
@@ -220,23 +225,40 @@ fn mutex_kind_from_static_initializer<'tcx>(
             },
         _ => {}
     }
-    throw_unsup_format!("unsupported static initializer used for `pthread_mutex_t`");
+    throw_ub_format!(
+        "`pthread_mutex_t` was not properly initialized at this location, or it got overwritten"
+    );
 }
 
 // # pthread_rwlock_t
 // We store some data directly inside the type, ignoring the platform layout:
-// - init: u32
+// - init: u8
 
 #[derive(Debug, Clone)]
 struct PthreadRwLock {
     rwlock_ref: RwLockRef,
 }
 
+impl SyncObj for PthreadRwLock {
+    fn on_access<'tcx>(&self, access_kind: AccessKind) -> InterpResult<'tcx> {
+        if !self.rwlock_ref.queue_is_empty() {
+            throw_ub_format!(
+                "{access_kind} of `pthread_rwlock_t` is forbidden while the queue is non-empty"
+            );
+        }
+        interp_ok(())
+    }
+
+    fn delete_on_write(&self) -> bool {
+        true
+    }
+}
+
 fn rwlock_init_offset<'tcx>(ecx: &MiriInterpCx<'tcx>) -> InterpResult<'tcx, Size> {
-    let offset = match &*ecx.tcx.sess.target.os {
-        "linux" | "illumos" | "solaris" | "freebsd" | "android" => 0,
+    let offset = match &ecx.tcx.sess.target.os {
+        Os::Linux | Os::Illumos | Os::Solaris | Os::FreeBsd | Os::Android => 0,
         // macOS stores a signature in the first bytes, so we move to offset 4.
-        "macos" => 4,
+        Os::MacOs => 4,
         os => throw_unsup_format!("`pthread_rwlock` is not supported on {os}"),
     };
     let offset = Size::from_bytes(offset);
@@ -245,11 +267,11 @@ fn rwlock_init_offset<'tcx>(ecx: &MiriInterpCx<'tcx>) -> InterpResult<'tcx, Size
     // the `init` field must start out not equal to LAZY_INIT_COOKIE.
     if !ecx.machine.pthread_rwlock_sanity.replace(true) {
         let static_initializer = ecx.eval_path(&["libc", "PTHREAD_RWLOCK_INITIALIZER"]);
-        let init_field = static_initializer.offset(offset, ecx.machine.layouts.u32, ecx).unwrap();
-        let init = ecx.read_scalar(&init_field).unwrap().to_u32().unwrap();
-        assert_ne!(
-            init, LAZY_INIT_COOKIE,
-            "PTHREAD_RWLOCK_INITIALIZER is incompatible with our initialization cookie"
+        let init_field = static_initializer.offset(offset, ecx.machine.layouts.u8, ecx).unwrap();
+        let init = ecx.read_scalar(&init_field).unwrap().to_u8().unwrap();
+        assert_eq!(
+            init, PTHREAD_UNINIT,
+            "PTHREAD_RWLOCK_INITIALIZER is incompatible with our initialization logic"
         );
     }
 
@@ -264,17 +286,20 @@ where
     'tcx: 'a,
 {
     let rwlock = ecx.deref_pointer_as(rwlock_ptr, ecx.libc_ty_layout("pthread_rwlock_t"))?;
-    ecx.lazy_sync_get_data(
+    ecx.get_immovable_sync_with_static_init(
         &rwlock,
         rwlock_init_offset(ecx)?,
-        || throw_ub_format!("`pthread_rwlock_t` can't be moved after first use"),
+        PTHREAD_UNINIT,
+        PTHREAD_INIT,
         |ecx| {
-            if !bytewise_equal_atomic_relaxed(
+            if !bytewise_equal(
                 ecx,
                 &rwlock,
                 &ecx.eval_path(&["libc", "PTHREAD_RWLOCK_INITIALIZER"]),
             )? {
-                throw_unsup_format!("unsupported static initializer used for `pthread_rwlock_t`");
+                throw_ub_format!(
+                    "`pthread_rwlock_t` was not properly initialized at this location, or it got overwritten"
+                );
             }
             interp_ok(PthreadRwLock { rwlock_ref: RwLockRef::new() })
         },
@@ -287,8 +312,8 @@ where
 
 #[inline]
 fn condattr_clock_offset<'tcx>(ecx: &MiriInterpCx<'tcx>) -> InterpResult<'tcx, u64> {
-    interp_ok(match &*ecx.tcx.sess.target.os {
-        "linux" | "illumos" | "solaris" | "freebsd" | "android" => 0,
+    interp_ok(match &ecx.tcx.sess.target.os {
+        Os::Linux | Os::Illumos | Os::Solaris | Os::FreeBsd | Os::Android => 0,
         // macOS does not have a clock attribute.
         os => throw_unsup_format!("`pthread_condattr` clock field is not supported on {os}"),
     })
@@ -322,13 +347,13 @@ fn condattr_set_clock_id<'tcx>(
 
 // # pthread_cond_t
 // We store some data directly inside the type, ignoring the platform layout:
-// - init: u32
+// - init: u8
 
 fn cond_init_offset<'tcx>(ecx: &MiriInterpCx<'tcx>) -> InterpResult<'tcx, Size> {
-    let offset = match &*ecx.tcx.sess.target.os {
-        "linux" | "illumos" | "solaris" | "freebsd" | "android" => 0,
+    let offset = match &ecx.tcx.sess.target.os {
+        Os::Linux | Os::Illumos | Os::Solaris | Os::FreeBsd | Os::Android => 0,
         // macOS stores a signature in the first bytes, so we move to offset 4.
-        "macos" => 4,
+        Os::MacOs => 4,
         os => throw_unsup_format!("`pthread_cond` is not supported on {os}"),
     };
     let offset = Size::from_bytes(offset);
@@ -337,11 +362,11 @@ fn cond_init_offset<'tcx>(ecx: &MiriInterpCx<'tcx>) -> InterpResult<'tcx, Size>
     // the `init` field must start out not equal to LAZY_INIT_COOKIE.
     if !ecx.machine.pthread_condvar_sanity.replace(true) {
         let static_initializer = ecx.eval_path(&["libc", "PTHREAD_COND_INITIALIZER"]);
-        let init_field = static_initializer.offset(offset, ecx.machine.layouts.u32, ecx).unwrap();
-        let init = ecx.read_scalar(&init_field).unwrap().to_u32().unwrap();
-        assert_ne!(
-            init, LAZY_INIT_COOKIE,
-            "PTHREAD_COND_INITIALIZER is incompatible with our initialization cookie"
+        let init_field = static_initializer.offset(offset, ecx.machine.layouts.u8, ecx).unwrap();
+        let init = ecx.read_scalar(&init_field).unwrap().to_u8().unwrap();
+        assert_eq!(
+            init, PTHREAD_UNINIT,
+            "PTHREAD_COND_INITIALIZER is incompatible with our initialization logic"
         );
     }
 
@@ -354,6 +379,21 @@ struct PthreadCondvar {
     clock: TimeoutClock,
 }
 
+impl SyncObj for PthreadCondvar {
+    fn on_access<'tcx>(&self, access_kind: AccessKind) -> InterpResult<'tcx> {
+        if !self.condvar_ref.queue_is_empty() {
+            throw_ub_format!(
+                "{access_kind} of `pthread_cond_t` is forbidden while the queue is non-empty"
+            );
+        }
+        interp_ok(())
+    }
+
+    fn delete_on_write(&self) -> bool {
+        true
+    }
+}
+
 fn cond_create<'tcx>(
     ecx: &mut MiriInterpCx<'tcx>,
     cond_ptr: &OpTy<'tcx>,
@@ -361,7 +401,7 @@ fn cond_create<'tcx>(
 ) -> InterpResult<'tcx, PthreadCondvar> {
     let cond = ecx.deref_pointer_as(cond_ptr, ecx.libc_ty_layout("pthread_cond_t"))?;
     let data = PthreadCondvar { condvar_ref: CondvarRef::new(), clock };
-    ecx.lazy_sync_init(&cond, cond_init_offset(ecx)?, data.clone())?;
+    ecx.init_immovable_sync(&cond, cond_init_offset(ecx)?, PTHREAD_INIT, data.clone())?;
     interp_ok(data)
 }
 
@@ -373,17 +413,20 @@ where
     'tcx: 'a,
 {
     let cond = ecx.deref_pointer_as(cond_ptr, ecx.libc_ty_layout("pthread_cond_t"))?;
-    ecx.lazy_sync_get_data(
+    ecx.get_immovable_sync_with_static_init(
         &cond,
         cond_init_offset(ecx)?,
-        || throw_ub_format!("`pthread_cond_t` can't be moved after first use"),
+        PTHREAD_UNINIT,
+        PTHREAD_INIT,
         |ecx| {
-            if !bytewise_equal_atomic_relaxed(
+            if !bytewise_equal(
                 ecx,
                 &cond,
                 &ecx.eval_path(&["libc", "PTHREAD_COND_INITIALIZER"]),
             )? {
-                throw_unsup_format!("unsupported static initializer used for `pthread_cond_t`");
+                throw_ub_format!(
+                    "`pthread_cond_t` was not properly initialized at this location, or it got overwritten"
+                );
             }
             // This used the static initializer. The clock there is always CLOCK_REALTIME.
             interp_ok(PthreadCondvar {
@@ -498,14 +541,14 @@ pub trait EvalContextExt<'tcx>: crate::MiriInterpCxExt<'tcx> {
                     MutexKind::Normal => throw_machine_stop!(TerminationInfo::Deadlock),
                     MutexKind::ErrorCheck => this.eval_libc_i32("EDEADLK"),
                     MutexKind::Recursive => {
-                        this.mutex_lock(&mutex.mutex_ref);
+                        this.mutex_lock(&mutex.mutex_ref)?;
                         0
                     }
                 }
             }
         } else {
             // The mutex is unlocked. Let's lock it.
-            this.mutex_lock(&mutex.mutex_ref);
+            this.mutex_lock(&mutex.mutex_ref)?;
             0
         };
         this.write_scalar(Scalar::from_i32(ret), dest)?;
@@ -525,14 +568,14 @@ pub trait EvalContextExt<'tcx>: crate::MiriInterpCxExt<'tcx> {
                     MutexKind::Default | MutexKind::Normal | MutexKind::ErrorCheck =>
                         this.eval_libc_i32("EBUSY"),
                     MutexKind::Recursive => {
-                        this.mutex_lock(&mutex.mutex_ref);
+                        this.mutex_lock(&mutex.mutex_ref)?;
                         0
                     }
                 }
             }
         } else {
             // The mutex is unlocked. Let's lock it.
-            this.mutex_lock(&mutex.mutex_ref);
+            this.mutex_lock(&mutex.mutex_ref)?;
             0
         }))
     }
@@ -575,11 +618,11 @@ pub trait EvalContextExt<'tcx>: crate::MiriInterpCxExt<'tcx> {
             throw_ub_format!("destroyed a locked mutex");
         }
 
+        // This write also deletes the interpreter state for this mutex.
         // This might lead to false positives, see comment in pthread_mutexattr_destroy
-        this.write_uninit(
-            &this.deref_pointer_as(mutex_op, this.libc_ty_layout("pthread_mutex_t"))?,
-        )?;
-        // FIXME: delete interpreter state associated with this mutex.
+        let mutex_place =
+            this.deref_pointer_as(mutex_op, this.libc_ty_layout("pthread_mutex_t"))?;
+        this.write_uninit(&mutex_place)?;
 
         interp_ok(())
     }
@@ -600,7 +643,7 @@ pub trait EvalContextExt<'tcx>: crate::MiriInterpCxExt<'tcx> {
                 dest.clone(),
             );
         } else {
-            this.rwlock_reader_lock(&rwlock.rwlock_ref);
+            this.rwlock_reader_lock(&rwlock.rwlock_ref)?;
             this.write_null(dest)?;
         }
 
@@ -615,7 +658,7 @@ pub trait EvalContextExt<'tcx>: crate::MiriInterpCxExt<'tcx> {
         if rwlock.rwlock_ref.is_write_locked() {
             interp_ok(Scalar::from_i32(this.eval_libc_i32("EBUSY")))
         } else {
-            this.rwlock_reader_lock(&rwlock.rwlock_ref);
+            this.rwlock_reader_lock(&rwlock.rwlock_ref)?;
             interp_ok(Scalar::from_i32(0))
         }
     }
@@ -648,7 +691,7 @@ pub trait EvalContextExt<'tcx>: crate::MiriInterpCxExt<'tcx> {
                 dest.clone(),
             );
         } else {
-            this.rwlock_writer_lock(&rwlock.rwlock_ref);
+            this.rwlock_writer_lock(&rwlock.rwlock_ref)?;
             this.write_null(dest)?;
         }
 
@@ -663,7 +706,7 @@ pub trait EvalContextExt<'tcx>: crate::MiriInterpCxExt<'tcx> {
         if rwlock.rwlock_ref.is_locked() {
             interp_ok(Scalar::from_i32(this.eval_libc_i32("EBUSY")))
         } else {
-            this.rwlock_writer_lock(&rwlock.rwlock_ref);
+            this.rwlock_writer_lock(&rwlock.rwlock_ref)?;
             interp_ok(Scalar::from_i32(0))
         }
     }
@@ -693,11 +736,11 @@ pub trait EvalContextExt<'tcx>: crate::MiriInterpCxExt<'tcx> {
             throw_ub_format!("destroyed a locked rwlock");
         }
 
+        // This write also deletes the interpreter state for this rwlock.
         // This might lead to false positives, see comment in pthread_mutexattr_destroy
-        this.write_uninit(
-            &this.deref_pointer_as(rwlock_op, this.libc_ty_layout("pthread_rwlock_t"))?,
-        )?;
-        // FIXME: delete interpreter state associated with this rwlock.
+        let rwlock_place =
+            this.deref_pointer_as(rwlock_op, this.libc_ty_layout("pthread_rwlock_t"))?;
+        this.write_uninit(&rwlock_place)?;
 
         interp_ok(())
     }
@@ -706,7 +749,7 @@ pub trait EvalContextExt<'tcx>: crate::MiriInterpCxExt<'tcx> {
         let this = self.eval_context_mut();
 
         // no clock attribute on macOS
-        if this.tcx.sess.target.os != "macos" {
+        if this.tcx.sess.target.os != Os::MacOs {
             // The default value of the clock attribute shall refer to the system
             // clock.
             // https://pubs.opengroup.org/onlinepubs/9699919799/functions/pthread_condattr_setclock.html
@@ -756,7 +799,7 @@ pub trait EvalContextExt<'tcx>: crate::MiriInterpCxExt<'tcx> {
 
         // Destroying an uninit pthread_condattr is UB, so check to make sure it's not uninit.
         // There's no clock attribute on macOS.
-        if this.tcx.sess.target.os != "macos" {
+        if this.tcx.sess.target.os != Os::MacOs {
             condattr_get_clock_id(this, attr_op)?;
         }
 
@@ -778,7 +821,7 @@ pub trait EvalContextExt<'tcx>: crate::MiriInterpCxExt<'tcx> {
 
         let attr = this.read_pointer(attr_op)?;
         // Default clock if `attr` is null, and on macOS where there is no clock attribute.
-        let clock_id = if this.ptr_is_null(attr)? || this.tcx.sess.target.os == "macos" {
+        let clock_id = if this.ptr_is_null(attr)? || this.tcx.sess.target.os == Os::MacOs {
             this.eval_libc("CLOCK_REALTIME")
         } else {
             condattr_get_clock_id(this, attr_op)?
@@ -834,8 +877,9 @@ pub trait EvalContextExt<'tcx>: crate::MiriInterpCxExt<'tcx> {
         &mut self,
         cond_op: &OpTy<'tcx>,
         mutex_op: &OpTy<'tcx>,
-        abstime_op: &OpTy<'tcx>,
+        timeout_op: &OpTy<'tcx>,
         dest: &MPlaceTy<'tcx>,
+        macos_relative_np: bool,
     ) -> InterpResult<'tcx> {
         let this = self.eval_context_mut();
 
@@ -844,7 +888,7 @@ pub trait EvalContextExt<'tcx>: crate::MiriInterpCxExt<'tcx> {
 
         // Extract the timeout.
         let duration = match this
-            .read_timespec(&this.deref_pointer_as(abstime_op, this.libc_ty_layout("timespec"))?)?
+            .read_timespec(&this.deref_pointer_as(timeout_op, this.libc_ty_layout("timespec"))?)?
         {
             Some(duration) => duration,
             None => {
@@ -853,14 +897,23 @@ pub trait EvalContextExt<'tcx>: crate::MiriInterpCxExt<'tcx> {
                 return interp_ok(());
             }
         };
-        if data.clock == TimeoutClock::RealTime {
-            this.check_no_isolation("`pthread_cond_timedwait` with `CLOCK_REALTIME`")?;
-        }
+
+        let (clock, anchor) = if macos_relative_np {
+            // `pthread_cond_timedwait_relative_np` always measures time against the
+            // monotonic clock, regardless of the condvar clock.
+            (TimeoutClock::Monotonic, TimeoutAnchor::Relative)
+        } else {
+            if data.clock == TimeoutClock::RealTime {
+                this.check_no_isolation("`pthread_cond_timedwait` with `CLOCK_REALTIME`")?;
+            }
+
+            (data.clock, TimeoutAnchor::Absolute)
+        };
 
         this.condvar_wait(
             data.condvar_ref,
             mutex_ref,
-            Some((data.clock, TimeoutAnchor::Absolute, duration)),
+            Some((clock, anchor, duration)),
             Scalar::from_i32(0),
             this.eval_libc("ETIMEDOUT"), // retval_timeout
             dest.clone(),
@@ -875,13 +928,14 @@ pub trait EvalContextExt<'tcx>: crate::MiriInterpCxExt<'tcx> {
         // Reading the field also has the side-effect that we detect double-`destroy`
         // since we make the field uninit below.
         let condvar = &cond_get_data(this, cond_op)?.condvar_ref;
-        if condvar.is_awaited() {
+        if !condvar.queue_is_empty() {
             throw_ub_format!("destroying an awaited conditional variable");
         }
 
+        // This write also deletes the interpreter state for this mutex.
         // This might lead to false positives, see comment in pthread_mutexattr_destroy
-        this.write_uninit(&this.deref_pointer_as(cond_op, this.libc_ty_layout("pthread_cond_t"))?)?;
-        // FIXME: delete interpreter state associated with this condvar.
+        let cond_place = this.deref_pointer_as(cond_op, this.libc_ty_layout("pthread_cond_t"))?;
+        this.write_uninit(&cond_place)?;
 
         interp_ok(())
     }
